
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/wsproxy:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI WSProxy',
    'entry_points': {
        'backendai_cli_v10': [
            'wsproxy = ai.backend.wsproxy.cli.__main__:main',
            'wsproxy.start-server = ai.backend.wsproxy.server:main',
        ],
    },
    'install_requires': (
        'Jinja2~=3.1.6',
        'PyJWT~=2.10.1',
        'PyYAML~=6.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp_jinja2~=1.6',
        'aiohttp~=3.11.16',
        'aiomonitor~=0.7.0',
        'aiotools~=1.9.0',
        'attrs>=25.3',
        'backend.ai-common==25.6.12rc1',
        'backend.ai-logging==25.6.12rc1',
        'backend.ai-plugin==25.6.12rc1',
        'click~=8.1.7',
        'pydantic[email]~=2.11.3',
        'pyhumps~=3.8.0',
        'setproctitle~=1.3.5',
        'tomlkit~=0.13.2',
        'types-Jinja2',
        'types-PyYAML',
        'uvloop~=0.21; sys_platform != "Windows"',
        'yarl~=1.19.0',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI WSProxy

""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-wsproxy',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.wsproxy': (
            'VERSION',
        ),
        'ai.backend.wsproxy.templates': (
            '__init__.py',
            'error.jinja2',
        ),
    },
    'packages': (
        'ai.backend.wsproxy',
        'ai.backend.wsproxy.api',
        'ai.backend.wsproxy.cli',
        'ai.backend.wsproxy.proxy.backend',
        'ai.backend.wsproxy.proxy.frontend',
        'ai.backend.wsproxy.proxy.frontend.http',
        'ai.backend.wsproxy.templates',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '25.6.12rc1',
    'zip_safe': False,
})
